import { Rule, Tree, SchematicsException } from '@angular-devkit/schematics';
import { AddToModuleContext } from './add-to-module-context';
import { strings } from '@angular-devkit/core';
import * as ts from 'typescript';

import { ModuleOptions, buildRelativePath } from '../schematics-angular-utils/find-module';
import { addDeclarationToModule, addExportToModule, addProviderToModule } from '../schematics-angular-utils/ast-utils';
import { InsertChange } from '../schematics-angular-utils/change';

enum TypeComponent {
  ComponentType,
  DirectiveType,
  ServiceType,
  PipeType,
}

export function addDeclarationToNgModule(options: ModuleOptions, exports: boolean): Rule {
  return (host: Tree) => {
    addDeclaration(host, options);
    if (exports) {
      addExport(host, options, TypeComponent.ComponentType);
    }
    return host;
  };
}

export function addProviderToNgModule(options: ModuleOptions, exports: boolean): Rule {
  return (host: Tree) => {
    addProvider(host, options);
    if (exports) {
      addExport(host, options, TypeComponent.ServiceType);
    }
    return host;
  };
}


function createAddToModuleContext(host: Tree, options: ModuleOptions, type: TypeComponent): AddToModuleContext {

  let typeForImport: string = "";
  let typeForClass: string = "";

  switch (type) {
    case TypeComponent.ComponentType: 
      typeForImport = "component";
      typeForClass = "Component";
    break;
    case TypeComponent.ServiceType:
      typeForImport = "service";
      typeForClass = "Service";
    break;
    default:
  }

  const result = new AddToModuleContext();

  if (!options.module) {
    throw new SchematicsException(`Module not found.`);
  }

  // Reading the module file
  const text = host.read(options.module);

  if (text === null) {
    throw new SchematicsException(`File ${options.module} does not exist.`);
  }

  const sourceText = text.toString('utf-8');
  result.source = ts.createSourceFile(options.module, sourceText, ts.ScriptTarget.Latest, true);

  const componentPath = `/${options.sourceDir}/${options.path}/`
		      + strings.dasherize(options.name) + '/'
		      + strings.dasherize(options.name)
          + '.'
          + typeForImport;

  result.relativePath = buildRelativePath(options.module, componentPath);

  result.classifiedName = strings.classify(`${options.name + typeForClass}`);

  return result;

}

function addDeclaration(host: Tree, options: ModuleOptions) {

  const context = createAddToModuleContext(host, options, TypeComponent.ComponentType);
  const modulePath = options.module || '';

  const declarationChanges = addDeclarationToModule(context.source,
    modulePath,
      context.classifiedName,
      context.relativePath);

  const declarationRecorder = host.beginUpdate(modulePath);
  for (const change of declarationChanges) {
    if (change instanceof InsertChange) {
      declarationRecorder.insertLeft(change.pos, change.toAdd);
    }
  }
  host.commitUpdate(declarationRecorder);
}

function addProvider(host: Tree, options: ModuleOptions) {

  const context = createAddToModuleContext(host, options, TypeComponent.ServiceType);
  const modulePath = options.module || '';

  const declarationChanges = addProviderToModule(context.source,
    modulePath,
      context.classifiedName,
      context.relativePath);

  const declarationRecorder = host.beginUpdate(modulePath);
  for (const change of declarationChanges) {
    if (change instanceof InsertChange) {
      declarationRecorder.insertLeft(change.pos, change.toAdd);
    }
  }
  host.commitUpdate(declarationRecorder);
}

function addExport(host: Tree, options: ModuleOptions, type: TypeComponent) {
  const context = createAddToModuleContext(host, options, type);
  const modulePath = options.module || '';

  const exportChanges = addExportToModule(context.source,
      modulePath,
      context.classifiedName,
      context.relativePath);

  const exportRecorder = host.beginUpdate(modulePath);

  for (const change of exportChanges) {
    if (change instanceof InsertChange) {
      exportRecorder.insertLeft(change.pos, change.toAdd);
    }
  }
  host.commitUpdate(exportRecorder);
}